from matplotlib import projections
import numpy as np
import math
import matplotlib
from matplotlib import pyplot as plt


def ackley(x: np.array,
           switch: np.array = None,
           disturb = 0.03,
           reversed=False):
    """
            f(x)=\sum(x_i^2)
        """
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 2:
        x = x.squeeze()

    x = x / 1e4
    result = (-20 * np.exp(-0.2 * np.sqrt(np.inner(x, x) / x.size)) - np.exp(
        np.cos(2 * np.pi * x).sum() / x.size) + 20 + np.e)

    if reversed:
        result =  -1 * result

    return result

if __name__ == '__main__':
    x = np.linspace(-10, 10, 1000)
    y = np.linspace(-10, 10, 1000)
    xv, yv = np.meshgrid(x, y)
    m, n = xv.shape
    ret = np.zeros(shape=(m,n))
    for i in range(m):
        for j in range(n):
            a, b = xv[i][j], yv[i][j]
            ret[i][j] = ackley(np.array([a, b]))
    

    fig = plt.figure()
    ax = plt.axes(projection='3d')
    surf = ax.plot_surface(xv, yv, ret)
    plt.show()





def levy(x: np.array,
           switch: np.array = None,
           disturb = 0.03,
           reversed=False):
    """
        f(x)=\sum(x_i^2)
    """
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 2:
        x = x.squeeze()

    x = x / 1e4
    w = []
    for idx in range(0, len(x)):
        w.append(1 + (x[idx] - 1) / 4)
    w = np.array(w)

    term1 = (np.sin(np.pi * w[0])) ** 2

    term3 = (w[-1] - 1) ** 2 * (1 + (np.sin(2 * np.pi * w[-1])) ** 2)

    term2 = 0
    for idx in range(1, len(w)):
        wi = w[idx]
        new = (wi - 1) ** 2 * (1 + 10 * (np.sin(np.pi * wi + 1)) ** 2)
        term2 = term2 + new

    result = term1 + term2 + term3

    if reversed:
        result =  -1 * result

    return result

def rosenbrock(x: np.array,
               switch: np.array = None,
               disturb = 0.03,
               reversed=False):
    """
        f(x)=np.sum(100 * (x[1:] - x[:-1] ** 2) ** 2 + (x[:-1] - 1) ** 2)
    """
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 2:
        x = x.squeeze()

    x = x / 1e4
    result = np.sum(100 * (x[1:] - x[:-1] ** 2) ** 2 + (x[:-1] - 1) ** 2)

    if reversed:
        result =  -1 * result

    return result


def rastrigin(x: np.array,
               switch: np.array = None,
               disturb = 0.03,
               reversed=False):
    """
        f(x)=10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x))
    """
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 2:
        x = x.squeeze()

    x = x / 1e4
    result = 10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x))

    if reversed:
        result = -1 * result

    return result

def spherefunc(x: np.array, 
               switch: np.array = None, 
               disturb = 0.03,
               reversed=False):
    
    """
        f(x)=\sum(x_i^2)
    """
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 1:
        x = np.expand_dims(x,0)
        
    if switch is not None and switch.shape.__len__() == 1:
        switch = np.expand_dims(switch,0)

    if reversed:
        x = 100 - x
    
    result = np.array([np.sum(_x**2) for _x in x]) * (np.random.randn(x.shape[0]) * disturb + 1)
    
    if switch is not None:
        result += spherefunc(x = x[...,:switch.shape[1]] * switch, disturb = disturb)
    
    return result


def alpinefunc(x: np.array, 
               switch: np.array = None, 
               disturb = 0.03, 
               reversed=False):
    
    """
        f(x)=\sum_{i=1}^n{|x_isin(x_i)+0.1x_i|}
    """
    
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 1:
        x = np.expand_dims(x,0)
        
    if switch is not None and switch.shape.__len__() == 1:
        switch = np.expand_dims(switch,0)
        
    if reversed:
        x = 100 - x

    result = np.array([np.sum(abs(_x * np.sin(_x) + 0.1 * _x)) for _x in x]) * (np.random.randn(x.shape[0]) * disturb + 1)

    if switch is not None:
        result += alpinefunc(x = x[...,:switch.shape[1]] * switch, disturb = disturb)
    
    return result


def multifunc(x: np.array, 
              switch: np.array = None, 
              disturb = 0.03):
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 1:
        x = np.expand_dims(x,0)
        
    if switch is not None and switch.shape.__len__() == 1:
        switch = np.expand_dims(switch,0)
    
    targetA = alpinefunc(x,switch,disturb,True)
    targetB = alpinefunc(x,switch,disturb,False)
    
    return np.concatenate(
        (np.expand_dims(targetA,axis=1),np.expand_dims(targetB,axis=1)),axis=1)
    
    
def multiFuncAvg(x: np.array, 
              switch: np.array = None, 
              disturb = 0.03):
    assert x.shape.__len__() == 1 or x.shape.__len__() == 2
    if x.shape.__len__() == 1:
        x = np.expand_dims(x,0)
        
    if switch is not None and switch.shape.__len__() == 1:
        switch = np.expand_dims(switch,0)
    
    targetA = alpinefunc(x,switch,disturb,True)
    targetB = alpinefunc(x,switch,disturb,False)
    
    result = np.concatenate(
        (np.expand_dims(targetA,axis=1),np.expand_dims(targetB,axis=1)),axis=1)
    
    return np.mean(result,axis=1)


def plot_func(f):
    x1 = np.linspace(-20, 20)
    x2 = np.linspace(-20, 20)
    X1, X2 = np.meshgrid(x1, x2)
    X1 = X1 * 1e4
    X2 = X2 * 1e4

    Z = np.zeros((X1.shape[0], X2.shape[0]))
    for i in range(X1.shape[0]):
        for j in range(X2.shape[0]):
            Z[i, j] = f(np.array([X1[i, j], X2[i, j]]))

    import matplotlib.pyplot as plt
    from mpl_toolkits import mplot3d

    # plt.contour(X1, X2, Z)

    fig = plt.figure(figsize=[12, 8])
    ax = plt.axes(projection='3d')
    ax.plot_surface(X1, X2, Z, cmap='jet', alpha=0.8)

#plot_func(alpinefunc)
#print("done")

